/*
 * (C) Copyright 2010
 * Texas Instruments, <www.ti.com>
 *
 * Aneesh V <aneesh@ti.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <common.h>
#include <spl.h>
#include <asm/u-boot.h>
#include <asm/utils.h>
#include <mmc.h>
#include <fat.h>
#include <version.h>

DECLARE_GLOBAL_DATA_PTR;


/* CRC16 implementation acording to CCITT standards */

static const uint16_t crc16tab[256]= {
    0x0000,0x1021,0x2042,0x3063,0x4084,0x50a5,0x60c6,0x70e7,
    0x8108,0x9129,0xa14a,0xb16b,0xc18c,0xd1ad,0xe1ce,0xf1ef,
    0x1231,0x0210,0x3273,0x2252,0x52b5,0x4294,0x72f7,0x62d6,
    0x9339,0x8318,0xb37b,0xa35a,0xd3bd,0xc39c,0xf3ff,0xe3de,
    0x2462,0x3443,0x0420,0x1401,0x64e6,0x74c7,0x44a4,0x5485,
    0xa56a,0xb54b,0x8528,0x9509,0xe5ee,0xf5cf,0xc5ac,0xd58d,
    0x3653,0x2672,0x1611,0x0630,0x76d7,0x66f6,0x5695,0x46b4,
    0xb75b,0xa77a,0x9719,0x8738,0xf7df,0xe7fe,0xd79d,0xc7bc,
    0x48c4,0x58e5,0x6886,0x78a7,0x0840,0x1861,0x2802,0x3823,
    0xc9cc,0xd9ed,0xe98e,0xf9af,0x8948,0x9969,0xa90a,0xb92b,
    0x5af5,0x4ad4,0x7ab7,0x6a96,0x1a71,0x0a50,0x3a33,0x2a12,
    0xdbfd,0xcbdc,0xfbbf,0xeb9e,0x9b79,0x8b58,0xbb3b,0xab1a,
    0x6ca6,0x7c87,0x4ce4,0x5cc5,0x2c22,0x3c03,0x0c60,0x1c41,
    0xedae,0xfd8f,0xcdec,0xddcd,0xad2a,0xbd0b,0x8d68,0x9d49,
    0x7e97,0x6eb6,0x5ed5,0x4ef4,0x3e13,0x2e32,0x1e51,0x0e70,
    0xff9f,0xefbe,0xdfdd,0xcffc,0xbf1b,0xaf3a,0x9f59,0x8f78,
    0x9188,0x81a9,0xb1ca,0xa1eb,0xd10c,0xc12d,0xf14e,0xe16f,
    0x1080,0x00a1,0x30c2,0x20e3,0x5004,0x4025,0x7046,0x6067,
    0x83b9,0x9398,0xa3fb,0xb3da,0xc33d,0xd31c,0xe37f,0xf35e,
    0x02b1,0x1290,0x22f3,0x32d2,0x4235,0x5214,0x6277,0x7256,
    0xb5ea,0xa5cb,0x95a8,0x8589,0xf56e,0xe54f,0xd52c,0xc50d,
    0x34e2,0x24c3,0x14a0,0x0481,0x7466,0x6447,0x5424,0x4405,
    0xa7db,0xb7fa,0x8799,0x97b8,0xe75f,0xf77e,0xc71d,0xd73c,
    0x26d3,0x36f2,0x0691,0x16b0,0x6657,0x7676,0x4615,0x5634,
    0xd94c,0xc96d,0xf90e,0xe92f,0x99c8,0x89e9,0xb98a,0xa9ab,
    0x5844,0x4865,0x7806,0x6827,0x18c0,0x08e1,0x3882,0x28a3,
    0xcb7d,0xdb5c,0xeb3f,0xfb1e,0x8bf9,0x9bd8,0xabbb,0xbb9a,
    0x4a75,0x5a54,0x6a37,0x7a16,0x0af1,0x1ad0,0x2ab3,0x3a92,
    0xfd2e,0xed0f,0xdd6c,0xcd4d,0xbdaa,0xad8b,0x9de8,0x8dc9,
    0x7c26,0x6c07,0x5c64,0x4c45,0x3ca2,0x2c83,0x1ce0,0x0cc1,
    0xef1f,0xff3e,0xcf5d,0xdf7c,0xaf9b,0xbfba,0x8fd9,0x9ff8,
    0x6e17,0x7e36,0x4e55,0x5e74,0x2e93,0x3eb2,0x0ed1,0x1ef0
};

uint16_t crc16(uint8_t const * buf_in, size_t len)
{
    size_t counter;
    uint8_t const * buf = (uint8_t const *)(buf_in);
    uint16_t crc = 0;
    for( counter = 0; counter < len; counter++)
        crc = (crc<<8) ^ crc16tab[((crc>>8) ^ *(char *)buf++)&0x00FF];
    return crc;
}

typedef struct { u32 cID; u32 addr; u32 regionSize; u32 payloadSize; u16 minVer; u16 maxVer; u16 CRC; u16 dummy; } flashmap_t;

u32 mmc_load_fgpa_raw(struct mmc *mmc)
{
    u16 CalculatedCRC = 0xffff, StoredFlashMapCRC = 0;
    u32 image_size_sectors, err, debugCount = 1000, i;
    u8* appBlockptr;
    ALLOC_CACHE_ALIGN_BUFFER(char, dst_buffer, 512);
    flashmap_t* flashMap = (flashmap_t*)(dst_buffer);
    flashmap_t  fpga, fpgab;

    /* read flash map data to find the image size & load address */
    err = mmc->block_dev.block_read(0, 0x900000 / 0x200, 1, (void *)dst_buffer);    // block 0x4800  = emmc adddress 0x900,000

    for (i = 0; i < 16; i++)
    {
        switch (flashMap[i].cID)
        {
            case 0x2:
                StoredFlashMapCRC = flashMap[i].CRC;
                flashMap[i].CRC = 0;
                CalculatedCRC = crc16((u8*)flashMap, 0x200);
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        flashMap[i].cID, flashMap[i].addr, flashMap[i].regionSize, flashMap[i].payloadSize, flashMap[i].minVer, flashMap[i].maxVer, flashMap[i].CRC);
                break;

            case 0x3:
                memcpy((u8*)&fpga, &flashMap[i], sizeof(flashmap_t));
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        fpga.cID, fpga.addr, fpga.regionSize, fpga.payloadSize, fpga.minVer, fpga.maxVer, fpga.CRC);
                break;

            case 0x4:
                memcpy((u8*)&fpgab, &flashMap[i], sizeof(flashmap_t));
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        fpgab.cID, fpgab.addr, fpgab.regionSize, fpgab.payloadSize, fpgab.minVer, fpgab.maxVer, fpgab.CRC);
                break;
 
            default:
                    // all others not required
                    break;
        }
    }

    /* If flash map invalid then use defaults */
    if ((fpga.payloadSize == 0) || (StoredFlashMapCRC != CalculatedCRC))
    {
        printf("MAP: Invalid - using fpga defaults Calc CRC:0x%x\n", CalculatedCRC);
        fpga.addr = 0x2000000;          /* emmc address */
        fpga.payloadSize = 0x800000;
    }

    image_size_sectors = fpga.payloadSize? fpga.payloadSize: fpga.regionSize;          // 8MBytes
    image_size_sectors = (image_size_sectors / 0x200) + 1;
    if (image_size_sectors > 0x4000) image_size_sectors = 0x4000;

    printf("Loading FPGA blkaddr:0x%x blksize:%u) [", fpga.addr / 0x200,  image_size_sectors);
    for(i = 0; i < image_size_sectors; i++)
    {
        appBlockptr = 0x100000 + (i*512);  /* SDRAM address */
        err = mmc->block_dev.block_read(0, (fpga.addr / 0x200) + i, 1, (void *)dst_buffer);
        if (err == 0) printf("mem read FPGA err %u\n", err);
        memcpy(appBlockptr, dst_buffer, 512);
        if (++debugCount > 1000)
        {
            printf(".");
            debugCount = 0;
        }
    }
    printf("]\n");

    return fpga.payloadSize;
}



static void mmc_load_image_raw(struct mmc *mmc)
{
    u16 CalculatedCRC = 0xffff, StoredFlashMapCRC = 0;
    u32 image_size_sectors, err, debugCount = 1000, i;
    u8* appBlockptr;
	ALLOC_CACHE_ALIGN_BUFFER(char, dst_buffer, 512);
    flashmap_t* flashMap = (flashmap_t*)(dst_buffer);
    flashmap_t  core1, core1b, core2, core2b;

	/* read flash map data to find the image size & load address */
	err = mmc->block_dev.block_read(0, 0x900000 / 0x200, 1, (void *)dst_buffer);    // block 0x4800  = emmc adddress 0x900,000

	for (i  = 0; i < 16; i++)
	{
	    switch (flashMap[i].cID)
	    {
	        case 0x2:
	            StoredFlashMapCRC = flashMap[i].CRC;
                flashMap[i].CRC = 0;
                CalculatedCRC = crc16((u8*)flashMap, 0x200);
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        flashMap[i].cID, flashMap[i].addr, flashMap[i].regionSize, flashMap[i].payloadSize, flashMap[i].minVer, flashMap[i].maxVer, flashMap[i].CRC);
                break;

	        case 0x5:
	            memcpy((u8*)&core1, &flashMap[i], sizeof(flashmap_t));
	            printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
	                    core1.cID, core1.addr, core1.regionSize, core1.payloadSize, core1.minVer, core1.maxVer, core1.CRC);
	            break;

            case 0x6:
                memcpy((u8*)&core1b, &flashMap[i], sizeof(flashmap_t));
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        core1b.cID, core1b.addr, core1b.regionSize, core1b.payloadSize, core1b.minVer, core1b.maxVer, core1b.CRC);
                break;

            case 0xd:
                memcpy((u8*)&core2, &flashMap[i], sizeof(flashmap_t));
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        core2.cID, core2.addr, core2.regionSize, core2.payloadSize, core2.minVer, core2.maxVer, core2.CRC);
                break;

            case 0xe:
                memcpy((u8*)&core2b, &flashMap[i], sizeof(flashmap_t));
                printf("MAP: [i:0x%x addr:0x%x rSize:0x%x pSize:0x%x min:%u max:%u crc:0x%x]\n",
                        core2b.cID, core2b.addr, core2b.regionSize, core2b.payloadSize, core2b.minVer, core2b.maxVer, core2b.CRC);
                break;

            default:
                    // all others not required
                    break;
	    }
	}

	/* If flash map invalid then use defaults */
	if ((StoredFlashMapCRC == 0) || (StoredFlashMapCRC != CalculatedCRC))
	{
	    printf("MAP: Invalid - using defaults Calc CRC:0x%x\n", CalculatedCRC);
	    core1.addr = 0x1000000;
	    core1.payloadSize = 0x800000;
        core1b.addr = 0x1800000;
        core1b.payloadSize = 0x800000;
        core2.payloadSize = 0;
        core2b.payloadSize = 0;
	}

	/* Ensure payload size isn't zero, as this crashes the CRC code */
	if (core1.payloadSize == 0) core1.payloadSize = core1.regionSize;

    image_size_sectors = core1.payloadSize? core1.payloadSize: core1.regionSize;          // 8MBytes
    image_size_sectors = (image_size_sectors / 0x200) + 1;
    if (image_size_sectors > 0x4000) image_size_sectors = 0x4000;
	
    printf("Loading Core-1 blkaddr:0x%x blksize:%u) [", core1.addr / 0x200,  image_size_sectors);
    for(i = 0; i < image_size_sectors; i++)
    {
        appBlockptr = 0x100000 + (i*512);
        err = mmc->block_dev.block_read(0, (core1.addr / 0x200) + i, 1, (void *)dst_buffer);
        if (err == 0) printf("mem read MainApp_Core1 err %u\n", err);
        memcpy(appBlockptr, dst_buffer, 512);
        if (++debugCount > 1000)
        {
        	printf(".");
        	debugCount = 0;
        }
    }

    printf("]\n");

	CalculatedCRC = crc16((u8*)0x100000, core1.payloadSize );
	printf("  CalcCRC:0x%x   core1.CRC:0x%x  core1.payloadSize:%u\n", CalculatedCRC, core1.CRC, core1.payloadSize );

	if ((CalculatedCRC != core1.CRC) || (core1.CRC == 0))
	{
	    printf("Loading Core-1 Backup blkaddr:0x%x blksize:%u) [", core1b.addr / 0x200, image_size_sectors);
	    for(i = 0; i < image_size_sectors; i++)
	    {
	        appBlockptr = 0x100000 + (i*512);
	        err = mmc->block_dev.block_read(0, (core1b.addr / 0x200) + i, 1, (void *)dst_buffer);
	        if (err == 0) printf("mem read MainApp_Core1 backup err %u\n", err);
	        memcpy(appBlockptr, dst_buffer, 512);
	        if (++debugCount > 1000)
	        {
	            printf(".");
	            debugCount = 0;
	        }
	    }
	    printf("]\n");

	    CalculatedCRC = crc16((u8*)0x100000, core1b.payloadSize );
	    printf("backup  CalcCRC:0x%x   core1b.CRC:0x%x  core1b.payloadSize:%u \n", CalculatedCRC, core1b.CRC, core1b.payloadSize );

	}

	/* CORE 2 LOADING */
	if (core2.payloadSize != 0)       /* Don't load if flash map invalid */
	{
        image_size_sectors = core2.payloadSize? core2.payloadSize: core2.regionSize;          // 8MBytes
        image_size_sectors = (image_size_sectors / 0x200) + 1;
        if (image_size_sectors > 0x4000) image_size_sectors = 0x4000;

        printf("Loading Core-2 blkaddr:0x%x blksize:%u) [", core2.addr / 0x200,  image_size_sectors);
        for(i = 0; i < image_size_sectors; i++)
        {
            appBlockptr = 0x10000000 + (i*512);
            err = mmc->block_dev.block_read(0, (core2.addr / 0x200) + i, 1, (void *)dst_buffer);
            if (err == 0) printf("mem read MainApp_Core2 err %u\n", err);
            memcpy(appBlockptr, dst_buffer, 512);
            if (++debugCount > 1000)
            {
                printf(".");
                debugCount = 0;
            }
        }
        printf("]\n");


        /* Check if primary core2 is ok, if not then load core2 backup */
        CalculatedCRC = crc16((u8*)0x10000000, core2.payloadSize );
        printf("  CalcCRC:0x%x   core2.CRC:0x%x  core2.payloadSize:%u\n", CalculatedCRC, core2.CRC, core2.payloadSize );
        if ((CalculatedCRC != core2.CRC) || (core2.CRC == 0))
        {
            image_size_sectors = core2b.payloadSize? core2b.payloadSize: core2b.regionSize;          // 8MBytes
            image_size_sectors = (image_size_sectors / 0x200) + 1;
            if (image_size_sectors > 0x4000) image_size_sectors = 0x4000;

            printf("Loading Core-2 backup blkaddr:0x%x blksize:%u) [", core2.addr / 0x200,  image_size_sectors);
            for(i = 0; i < image_size_sectors; i++)
            {
                appBlockptr = 0x10000000 + (i*512);
                err = mmc->block_dev.block_read(0, (core2.addr / 0x200) + i, 1, (void *)dst_buffer);
                if (err == 0) printf("mem read Core2 backup err %u\n", err);
                memcpy(appBlockptr, dst_buffer, 512);
                if (++debugCount > 1000)
                {
                    printf(".");
                    debugCount = 0;
                }
            }
            printf("]\n");
        }
	}
}

static void mmc_load_image_mbr(struct mmc *mmc)
{
#define MBR_SIGNATURE				(0xAA55)
#define MBR_SIGNATURE_OFFSET			(0x1FE)
#define MBR_NUM_OF_PARTITIONS			(4)
#define MBR_PARTITION_OFFSET			(0x1BE)
#define MBR_PARTITION_TYPE_OFFSET		(4)
#define MBR_PARTITION_START_SECTOR_OFFSET	(8)
#define MBR_PARTITION_ENTRY_SIZE		(16)
#define CONFIG_SYS_MMCSD_MBR_MODE_U_BOOT_PARTITION_ID	(0xA2)

	u32 i;
	u32 part_type;
	u32 offset = 0;
	u16 sign;
	u8 *p_entry;
	u32 image_size_sectors, err;
	const struct image_header *header;

	header = (struct image_header *)(CONFIG_SYS_TEXT_BASE -
						sizeof(struct image_header));

	/* search the MBR to get the parition offset */
	/* Read MBR */
	err = mmc->block_dev.block_read(0, 0, 1, (void *)header);
	if (err <= 0)
		goto end;
	memcpy(&sign, (u8 *)header + MBR_SIGNATURE_OFFSET, sizeof(sign));
	if (sign != MBR_SIGNATURE) {
		puts("spl: No MBR signature is found\n");
		hang();
	}

	/* Lookup each partition entry for partition ID. */
	p_entry = (u8 *)header + MBR_PARTITION_OFFSET;
	for (i = 0; i < MBR_NUM_OF_PARTITIONS; i++) {
		part_type = p_entry[MBR_PARTITION_TYPE_OFFSET];

		if (part_type ==
			CONFIG_SYS_MMCSD_MBR_MODE_U_BOOT_PARTITION_ID){
			/* Partition found */
			memcpy(&offset,
				p_entry + MBR_PARTITION_START_SECTOR_OFFSET,
				sizeof(offset));
			debug("spl: Partition offset 0x%x sectors\n", offset);
			break;
		}
		p_entry += MBR_PARTITION_ENTRY_SIZE;
	}


	/* read image header to find the image size & load address */
	err = mmc->block_dev.block_read(0,
			(offset + CONFIG_SYS_MMCSD_RAW_MODE_U_BOOT_SECTOR), 1,
			(void *)header);

	if (err <= 0)
		goto end;

	spl_parse_image_header(header);

	/* convert size to sectors - round up */
	image_size_sectors = (spl_image.size + mmc->read_bl_len - 1) /
				mmc->read_bl_len;

	/* Read the header too to avoid extra memcpy */
	err = mmc->block_dev.block_read(0,
			(offset + CONFIG_SYS_MMCSD_RAW_MODE_U_BOOT_SECTOR),
			image_size_sectors, (void *)spl_image.load_addr);

end:
	if (err <= 0) {
		printf("spl: mmc blk read err - %d\n", err);
		hang();
	}
}

#ifdef CONFIG_SPL_FAT_SUPPORT
static void mmc_load_image_fat(struct mmc *mmc)
{
	s32 err;
	struct image_header *header;

	header = (struct image_header *)(CONFIG_SYS_TEXT_BASE -
						sizeof(struct image_header));

	err = fat_register_device(&mmc->block_dev,
				CONFIG_SYS_MMC_SD_FAT_BOOT_PARTITION);
	if (err) {
		printf("spl: fat register err - %d\n", err);
		hang();
	}

	err = file_fat_read(CONFIG_SPL_FAT_LOAD_PAYLOAD_NAME,
				(u8 *)header, sizeof(struct image_header));
	if (err <= 0)
		goto end;

	spl_parse_image_header(header);

	err = file_fat_read(CONFIG_SPL_FAT_LOAD_PAYLOAD_NAME,
				(u8 *)spl_image.load_addr, 0);

end:
	if (err <= 0) {
		printf("spl: error reading image %s, err - %d\n",
			CONFIG_SPL_FAT_LOAD_PAYLOAD_NAME, err);
		hang();
	}
}
#endif

void spl_mmc_load_image(void)
{
	struct mmc *mmc;
	int err;
	u32 boot_mode = MMCSD_MODE_RAW;

	mmc_initialize(gd->bd);
	/* We register only one device. So, the dev id is always 0 */
	mmc = find_mmc_device(0);
	if (!mmc) {
		puts("spl: mmc device not found!!\n");
		hang();
	}

	err = mmc_init(mmc);
	if (err) {
		printf("spl: mmc init failed: err - %d\n", err);
		hang();
	}
	/*boot_mode = spl_boot_mode();*/
	if (boot_mode == MMCSD_MODE_RAW) {
		debug("boot mode - RAW\n");
		mmc_load_image_raw(mmc);
	} else if (boot_mode == MMCSD_MODE_MBR) {
		debug("boot mode - MBR\n");
		mmc_load_image_mbr(mmc);
#ifdef CONFIG_SPL_FAT_SUPPORT
	} else if (boot_mode == MMCSD_MODE_FAT) {
		debug("boot mode - FAT\n");
		mmc_load_image_fat(mmc);
#endif
	} else {
		puts("spl: wrong MMC boot mode\n");
		hang();
	}
}
